! WHIZARD 1.11 May 16 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module events

  use kinds, only: default !NODEP!
  use limits, only: BUFFER_SIZE
  use constants
  use lorentz
  use parameters, only: parameter_set !NODEP!
!  use polarization
  use particles
  use jetset_interface !NODEP!
  use stdhep_interface !NODEP!

  implicit none
  private

  integer, parameter, public :: SCREEN_FMT=0, HEPEVT_FMT=1, RAW_FMT=-1
  integer, parameter, public :: SHORT_FMT=2
  integer, parameter, public :: STDHEP_FMT=3
  integer, parameter, public :: JETSET_FMT1=11, JETSET_FMT2=12, JETSET_FMT3=13

  type, public :: event
     integer :: count, count_internal, n_in, n_out, n_tot
     logical :: scattering
     real(kind=default) :: sqrts, sqrts_hat, s_hat, flux_factor
     real(kind=default), dimension(2) :: x_parton
     type(particle), dimension(:), pointer :: prt
     real(kind=default) :: mc_function_value
     real(kind=default), dimension(:,:), pointer :: rho_out
!     type(spin_density_vector) :: rho_in, rho_out
!     type(spin_2transition_matrix) :: T_fi
     integer, dimension(:), pointer :: hel_out
     type(lorentz_transformation) :: L0
  end type event

  public :: create, destroy
  public :: read, write
  public :: write_process
  public :: set_in_state
  public :: set_out_state
  public :: get
  public :: merge

  interface create
     module procedure event_create
  end interface
  interface destroy
     module procedure event_destroy
  end interface
  interface write
     module procedure event_write_unit
  end interface
  interface read
     module procedure event_read_unit
  end interface
  interface write_process
     module procedure write_process_unit
  end interface
  interface set_in_state
     module procedure event_in_state_sqrts
  end interface
  interface set_in_state
     module procedure event_in_state_prt
  end interface
  interface set_out_state
     module procedure event_out_state_codes
  end interface
  interface get
     module procedure event_get_particle
  end interface
  interface merge
     module procedure event_apply_decay
  end interface

contains

  subroutine event_create(evt, n_in, n_out, n_hel_out, n_flv_out)
    type(event), intent(out) :: evt
    integer, intent(in) :: n_in, n_out, n_flv_out, n_hel_out
!    integer, dimension(:), intent(in) :: code
!    integer :: i
!    integer, dimension(n_in) :: spin_in
!    integer, dimension(n_out) :: spin_out
    !allocate(evt)
    evt%count = 0
    evt%count_internal = 0
    evt%scattering = (n_in==2)
    evt%n_in = n_in
    evt%n_out = n_out
    evt%n_tot = n_in + n_out
    if (evt%scattering) then
       allocate(evt%prt(-n_in:n_out))
    else
       allocate(evt%prt(0:n_out))
    end if
    allocate (evt%rho_out(n_hel_out, n_flv_out))
!     do i=1, n_in
!        spin_in(i) = particle_spin_type(code(i))
!     end do
!     call create(evt%rho_in, spin_in)
!     do i=1, n_out
!        spin_out(i) = particle_spin_type(code(n_in+i))
!     end do
!     call create(evt%rho_out, spin_out)
!     call create(evt%T_fi, spin_out, spin_in)
    evt%rho_out = 0
    evt%mc_function_value = 0
    allocate(evt%hel_out(n_out))
    evt%hel_out = 0
  end subroutine event_create

  subroutine event_destroy(evt)
    type(event), intent(inout) :: evt
    deallocate(evt%hel_out)
    deallocate(evt%rho_out)
!     call destroy(evt%T_fi)
!     call destroy(evt%rho_in)
!     call destroy(evt%rho_out)
    deallocate(evt%prt)
    !deallocate(evt)
  end subroutine event_destroy

  subroutine event_write_unit (u, evt, fmt, probability, weight, channel)
    integer, intent(in) :: u
    type(event), intent(in) :: evt
    integer, intent(in), optional :: fmt
    real(kind=default), intent(in), optional :: probability, weight
    integer, intent(in), optional :: channel
    integer :: mode, i, j
    real(kind=default), dimension(0:3) :: p
    mode = SCREEN_FMT;  if (present(fmt)) mode = fmt
    select case(mode)
    case(SCREEN_FMT)
       write(u,*) '============================================================================'
       write(u,*) 'Event # ', evt%count
       write(u,*) 'count =      ', evt%count_internal
       write(u,*) 'sqrts =      ', evt%sqrts
       write(u,*) 'sqrts_hat =  ', evt%sqrts_hat
       write(u,*) 'flux factor =', evt%flux_factor
       if (evt%scattering) write(u,*) 'x_parton =', evt%x_parton
       write(u,*)
       write(u,*) 'In-state:'
       call write(u,evt%prt(0))
       if (evt%scattering) then
          do i=1,2
             call write(u,evt%prt(-i))
          end do
       end if
       ! call write(u, evt%rho_in)
       ! write(u,*)
       ! call write(u, evt%T_fi)
       write(u,*)
       write(u,*) 'Out-state:'
       do i=1,evt%n_out
          call write(u,evt%prt(i))
       end do
       ! call write(u, evt%rho_out)
       ! write(u,*) 'Squared matrix element:', evt%sqme
       write(u,*) 'MC function value:', evt%mc_function_value
       write(u,*) 'Helicity vector:', evt%hel_out
    case(HEPEVT_FMT)
       write(u,11) evt%n_out
       do i=1, evt%n_out
          p = array(particle_momentum(evt%prt(i)))
          write (u,13) 1, particle_code(evt%prt(i)), &
               & (0, j=1,2), (0, j=1,2), evt%hel_out(i)
          write (u,12) p(1:3), p(0), particle_mass(evt%prt(i))
          write (u,12) (0._default, j=1,5)
       end do
    case(SHORT_FMT)
       write(u,11) evt%n_out
       do i=1, evt%n_out
          p = array(particle_momentum(evt%prt(i)))
          write (u,13) particle_code(evt%prt(i)), evt%hel_out(i)
          write (u,12) p(1:3), p(0), particle_mass(evt%prt(i))
       end do
    case(RAW_FMT)
       j=0;  if (evt%scattering) j=2
       do i=0,j
          write(u)  particle_code(evt%prt(-i)), &
               & array(particle_momentum(evt%prt(-i))), &
               & particle_mass(evt%prt(-i))
       end do
       do i=1, evt%n_out
          write(u) particle_code(evt%prt(i)), &
               & array(particle_momentum(evt%prt(i))), &
               & particle_mass(evt%prt(i)), &
               & evt%hel_out(i)
       end do
       if (present(probability).and.present(weight).and.present(channel)) &
            & write(u) probability, weight, channel
    case(STDHEP_FMT)
       call stdhep_write
    case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
       call jetset_write (mode-10)
    case default
       stop "Invalid event output format"
    end select
  11 format(1I5)
  12 format(10F17.10)
  13 format(I9,I9,5I5)
  end subroutine event_write_unit

  subroutine event_read_unit(u, evt, fmt, probability, weight, channel, iostat)
    integer, intent(in) :: u
    type(event), intent(inout) :: evt
    integer, intent(in), optional :: fmt
    real(kind=default), intent(out), optional :: probability, weight
    integer, intent(out), optional :: channel, iostat
    integer :: mode, i, j, code, idummy
    real(kind=default) :: mass, dummy
    real(kind=default), dimension(0:3) :: p
    mode = SCREEN_FMT;  if (present(fmt)) mode = fmt
    select case(mode)
    case(RAW_FMT)
       iostat = 0
       j=0;  if (evt%scattering) j=2
       do i=0,j
          read(u, iostat=iostat) code, p, mass
          if (iostat/=0) return
          evt%prt(-i) = particle_new(code, mass, &
               & four_momentum_moving(p(0), three_momentum_moving(p(1:3))))
       end do
       do i=1, evt%n_out
          read(u, iostat=iostat) code, p, mass, evt%hel_out(i)
          evt%prt(i) = particle_new(code, mass, &
               & four_momentum_moving(p(0), three_momentum_moving(p(1:3))))
       end do
       if (present(probability).and.present(weight).and.present(channel)) &
            & read(u) probability, weight, channel
    case default
       stop "Can read events in raw format only"
    end select
  11 format(1I5)
  12 format(10F17.10)
  13 format(I9,I9,5I5)
  end subroutine event_read_unit

  subroutine write_process_unit (u, process_id, code, n_in, n_out)
    integer, intent(in) :: u
    character(len=*), intent(in) :: process_id
    integer, dimension(:), intent(in) :: code
    integer, intent(in) :: n_in, n_out
    integer :: i, len
    character(len=BUFFER_SIZE), dimension(0:size(code)+1) :: buffer
    write(u,'(1x,A,1x,A,A,2x)') "! Process", trim(process_id), ":"
    write(u,'(1x,A)', advance='no') "!"
    write(buffer(0), '(1x,A)') "!"
    do i=1, n_in
       call write_prt (u, buffer(i), code(i), ibset(0,n_in+n_out-i))
    end do
    write(u,'(2x,A2)', advance='no') "->"
    write(buffer(n_in+1),'(2x,A2)') "->"
    do i=1, n_out
       call write_prt (u, buffer(n_in+i+1), code(n_in+i), ibset(0,i-1))
    end do
    write(u,*)
    do i=0, n_in+n_out+1
       write(u,'(A)', advance='no') trim(buffer(i))
    end do
    write(u,*)
  contains
    subroutine write_prt (u, buffer, code, bincode)
      integer, intent(in) :: u, code, bincode
      character(len=*), intent(inout) :: buffer
      character(len=8) :: fmt
      len = len_trim(particle_name(code))
      if (len<3) len = 3
      write(fmt, '(A5,I1,A2)') "(1x,A", len, ")"
      write(u, fmt, advance='no') trim(particle_name(code))
      write(fmt, '(A5,I1,A2)') "(1x,I", len, ")"
      write(buffer, fmt) bincode
    end subroutine write_prt
  end subroutine write_process_unit

  subroutine event_in_state_sqrts(evt, code, par, sqrts, xi, rapidity, ok)
    type(event), intent(inout) :: evt
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default), intent(in) :: sqrts
    real(kind=default), dimension(2), intent(in), optional :: xi
    real(kind=default), intent(in), optional :: rapidity
    logical, intent(out), optional :: ok
    integer :: root_code, i, n_in
    real(kind=default) :: pp, rap, sh, sqsh, lda
    real(kind=default), dimension(2) :: mass
    if (present(ok)) ok = .true.
    if (.not.present(xi)) then
       evt%x_parton = 1
       sqsh = sqrts
       sh   = sqrts**2
    else
       evt%x_parton = xi
       sh   = xi(1)*xi(2)*sqrts**2
       sqsh = sqrt(sh)
    end if
    evt%sqrts     = sqrts
    evt%sqrts_hat = sqsh
    evt%s_hat     = sh
    if (evt%scattering) then
       n_in = 2
       root_code = COMPOSITE
       do i = 1,2
          mass(i) = particle_mass(code(i), par)
       end do
       lda = lambda(evt%s_hat, mass(1)**2, mass(2)**2)
       if (lda <= 0) then
          if (present(ok)) then
             ok = .false.; return
          else
             stop 'Total energy too low for initial state'
          end if
       end if
       pp = sqrt(lda)
       evt%flux_factor = conv * twopi4 / (2*pp)
       evt%prt(-1) = particle_new( &
            & code = code(1),  mass = mass(1), &
            & p = four_momentum_moving( &
            &       (sqsh + (mass(1)**2-mass(2)**2)/sqsh)/2,  pp/(2*sqsh), 3) )
       evt%prt(-2) = particle_new( &
            & code = code(2),  mass = mass(2), &
            & p = four_momentum_moving( &
            &       (sqsh - (mass(1)**2-mass(2)**2)/sqsh)/2, -pp/(2*sqsh), 3) )
    else
       n_in = 0
       root_code = code(1)
       evt%flux_factor = conv * twopi4 / (2*sqsh)
    end if
    evt%prt(0) = particle_new( root_code, sqsh )
    if (present(rapidity).or.present(xi)) then
       rap = 0
       if (present(rapidity)) rap = rapidity
       if (present(xi)) then
          if (all(xi>0)) then
             rap = rap + log(xi(1)/xi(2))/2
          else
             if (present(ok)) then
                ok = .false.; return
             else
                stop 'Non-positive x value'
             end if
          end if
       end if
       evt%L0 = boost(sinh(rap),3)
       do i = 0, n_in
          evt%prt(-i) = evt%L0 * evt%prt(-i)
       end do
    else
       evt%L0 = unit_lorentz_transformation
    end if
  end subroutine event_in_state_sqrts

  subroutine event_in_state_prt(evt, prt, ok)
    type(event), intent(inout) :: evt
    type(particle), dimension(:), intent(in) :: prt
    logical, intent(out), optional :: ok
    integer :: i, n_in
    real(kind=default) :: lda
    type(four_momentum) :: p0
    if (present(ok)) ok = .true.
    if (evt%scattering) then
       n_in = 2
       p0 = particle_momentum(prt(1)) + particle_momentum(prt(2))
       evt%prt(0) = particle_new( code = COMPOSITE,  p = p0 )
       do i=1,2
          evt%prt(-i) = prt(i)
       end do
       lda = lambda(evt%s_hat, &
            &       particle_mass(prt(1))**2, particle_mass(prt(2))**2)
       if (lda<=0) then
          if (present(ok)) then
             ok = .false.; return
          else
             stop 'Total energy too low for initial state'
          end if
       end if
       evt%flux_factor = conv * twopi4 / (2 * sqrt(lda))
    else
       n_in = 0
       evt%prt(0) = prt(1)
       evt%flux_factor = conv * twopi4 / (2 * particle_mass(evt%prt(0)))
    end if
    evt%sqrts     = particle_mass(evt%prt(0))
    evt%sqrts_hat = evt%sqrts
    evt%s_hat     = evt%sqrts**2
    evt%L0        = boost( space_part(p0) / evt%sqrts )
    evt%x_parton  = 1
  end subroutine event_in_state_prt

  subroutine event_out_state_codes(evt, code, par)
    type(event), intent(inout) :: evt
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer :: i
    do i=1, evt%n_out
       evt%prt(i) = particle_new( code(i), par)
    end do
  end subroutine event_out_state_codes

  subroutine event_get_particle(evt, prt, code)
    type(event), intent(in) :: evt
    type(particle), intent(out) :: prt
    integer, intent(in) :: code
    integer :: i
    prt = particle_null
    do i=1, evt%n_out
       if (particle_code(evt%prt(i)) == code) then
          prt = evt%prt(i)
          exit
       end if
    end do
  end subroutine event_get_particle
  subroutine event_apply_decay(evt, evt_daughter)
    type(event), intent(inout) :: evt
    type(event), intent(in) :: evt_daughter
    type(particle), dimension(:), pointer :: prt
    integer :: i,n
    n = evt%n_out
    do i=1, n
       if (particle_code(evt%prt(i)) == particle_code(evt_daughter%prt(0))) &
            & then
          n = n + evt_daughter%n_out - 1
          if (evt%scattering) then
             allocate(prt(-2:n))
             prt(-2:-1) = evt%prt(-2:-1)
          else
             allocate(prt(0:n))
          end if
          prt(0:evt%n_out)  = evt%prt(0:)
          prt(i)                  = evt_daughter%prt(1)
          prt(evt%n_out+1:) = evt_daughter%prt(2:)
          deallocate(evt%prt)
          evt%prt => prt
          evt%n_out = n
          exit
       end if
    end do
  end subroutine event_apply_decay


end module events
